import os
import sys
from datetime import date
from selenium import webdriver
from webdriver_manager.chrome import ChromeDriverManager
from .selenium.common.exceptions import NoSuchElementException
from .DOM_elements import PageLocations


class MiningEngine:
    '''
    A simple mining engine to gather information from article-hosting websites.

    The locations in the DOM for each element are given by a PageLocations object
    in the constructor. This class then uses those instructions to gather the info.

    To give instructions to mine data for a location stored in :code:`PageLocations.xyz`,
    you should create/override the :code:`get_xyz()` function, like so:

    .. code-block:: python

        def get_xyz(self, element):
            \'\'\'
            Short description

            Arguments
            ---------
            element: :class:`centaurminer.Element`
                Location of the given element in the page.
           \'\'\'
           # Add instructions here, using self.wd and self.get
           # Below is equivalent to what happens without this override function
           return self.get(element)

    Arguments
    ---------
    site_locations : :class:`centaurMiner.PageLocations`
        A class reference to a subclass of :class:`centaurMiner.PageLocations`.
    driver_path : str, optional
        The location of your webdriver. If this is not specified, one will be installed/cached automatically.
    headless : bool, optional
        If False, the webdriver will open a GUI as it performs its tasks. Defaults to True (no GUI).

    Attributes
    ----------
    site : :class:`PageLocations <centaurminer.PageLocations>`
        Stores the location of the elements you want to extract information from.
    wd : :class:`selenium.webdriver`
        The webdriver used to connect to and collect DOM elements from the URL you specify.
    results : dict
        Storage dictionary for the results from data gathering.

    '''

    def __init__(self, site_locations: PageLocations, driver_path=None, headless=True):
        '''
        Arguments:
          site: Should be a class reference, not a class object. PageLocations are static classes, so you never need to instantiate them.
          driver_path: Path to the driver location on your machine.
          headless: If False, the webdriver will open a GUI while running.
        '''
        self.site = site_locations
        self.wd = self._init_selenium(driver_path, headless)
        self.results = {}

    def gather(self, url):
        '''
        Gather the information denoted in self.site from a single page.

        Arguments:

        * url: (string) URL for the site you want to mine data from.
        '''
        self.results = {}  # Reset for subsequent gathers
        self.wd.get(url)
        for info in self.site._elements():
            get_func = getattr(self, "get_" + info, None)
            element = getattr(self.site, info)

            # Check for strings - just store them directly
            if isinstance(element, str):
                self.results[info] = element
                continue

            if get_func is None and element.needsInstructions:
                print(self.site.__name__ + "." + info, "needs further instructions to process - add it to the engine.")
                continue
            elif get_func is None:
                get_func = self.get
            self.results[info] = get_func(element)
        self.results['url'] = url
        self.results['date_aquisition'] = date.today().strftime("%Y-%m-%d")

    def get_authors(self, element):
        '''
        Override for author collection instructions.

        Simply gets the list of authors and joins them together, using :func:`TagList`

        Arguments
        ---------
        element : :class:`centaurminer.Element`
            The location of the element to mine data from - in this case, it's several elements located with the same identifier.
        '''
        from .utils import TagList
        objs = self.get(element, several=True)
        return TagList(objs)

    def get(self, element, several=False):
        '''
        Default method for extracting an element from the page.

        Handles errors gracefully and waits for the element to become visible before grabbing it.
        When creating custom :code:`get_*****` functions, use this function to grab the data from the element, before doing additional processing on it.

        Arguments
        ---------
        element : :class:`centaurminer.Element`
            The location of the element to mine data from.
        several : Boolean
            Use to indicate that we should get all elements of this type, instead of the first one on the page.

            .. note::
                You should only use :code:`several=True` in a custom ``get_*`` method, so you can do more processing
                after getting this list of elements.

        '''
        if element is None:
            return None
        try:
            if several:
                pageObj = self.wd.find_elements(element.method, element.selector)
                return [self._extract(element, obj) for obj in pageObj]
            else:
                pageObj = self.wd.find_element(element.method, element.selector)
                return self._extract(element, pageObj)
        except NoSuchElementException:
            return None

    def _extract(self, element, obj):
        if element.attribute is None:
            return obj.text
        else:
            return obj.get_attribute(element.attribute)

    def _init_selenium(self, driver_path=None, headless=True):
        # Install a webdriver/get one from the cache if a path isn't given
        if driver_path is None:
            driver_path = ChromeDriverManager().install()

        print("Headless:", headless)
        print("driver path:", driver_path)
        chrome_options = webdriver.ChromeOptions()
        chrome_options.add_argument('--disable-dev-shm-usage')
        chrome_options.add_argument("--disable-logging")
        chrome_options.add_argument("log-level=3")
        chrome_options.add_argument('--no-sandbox')
        chrome_options.add_argument("--disable-gpu")
        chrome_options.add_argument("--kiosk")
        chrome_options.add_argument("--disable-extensions")

        if headless:
            chrome_options.add_argument('--headless')
        else:
            chrome_options.add_argument("--window-size=1920,1080")
            chrome_options.add_argument("--start-maximized")

        chrome_options.add_experimental_option
        (
            'prefs',
            {"download.default_directory": os.getcwd(),
                "download.prompt_for_download": False,
                "download.directory_upgrade": True,
                "plugins.always_open_pdf_externally": True
             }
        )
        return webdriver.Chrome(driver_path,
                                options=chrome_options)


if __name__ == "__main__":
    print(sys.path)
    miner = MiningEngine(PageLocations)
    miner.gather("https://www.thelancet.com/journals/lancet/article/PIIS0140-6736(20)31352-0/fulltext")
    print(miner.results)
    print(miner.results.keys())



from selenium.webdriver.common.by import By


class Element:
    '''
    Simple struct to store instructions to find an element on a page.

    Attributes
    ----------
    method : str
        The name of the method used to locate the element. Common values include "id", "xpath", and "css_selector" - look at `<https://selenium-python.readthedocs.io/locating-elements.html>`_ for more info. This is passed directly into the ``selenium.webdriver.common.By`` constructor.
    selector : str
        Combined with `method`, defines a way to find a specific element on the DOM. This is the actual xpath, css selector, etc., used to find the element.
    needsInstructions : Boolean
        If true, this element requires instructions in the :class:`centaurminer.MiningEngine` class to extract - used for the Complex subclass.
    '''
    def __init__(self, method, selector):
        self.needsInstructions = False
        self.attribute = None
        if isinstance(method, str):
            self.method = getattr(By, method.upper())
        else:
            self.method = method
            print(type(self.method))
        self.selector = selector

    def get_attribute(self, attributeName):
        '''
        Indicate which attribute of the element should be extracted.

        Arguments
        ---------
        attributeName : str
            Name of the attribute to be extracted.

        Returns
        -------
        Element
            Used for chaining with the contructor.
        '''
        self.attribute = attributeName
        return self


class MetaData(Element):
    '''
    A special type of Element that's derived from the metadata

    Arguments
    ---------
    name : str
        The 'name' or 'property' value of the Metadata this object points to.
    '''

    def __init__(self, name):
        super().__init__("XPATH", "html/head/meta[@name='" + name + "' or @property='" + name + "']")
        self.attribute = "content"


class Complex(Element):
    '''
    This represents an element that needs further directions to extract.

    An error will be thrown when trying to gather data, if the :class:`centaurminer.MiningEngine` doing the mining does not have an instruction for a Complex element.
    '''
    def __init__(self):
        self.needsInstructions = True


class PageLocations:
    '''
    The base class for locating article elements on a site.

    Some fields are gathered in a default way, based on standard metadata:

        * title : MetaData("citation_title")
        * authors : MetaData("citation_author")
        * doi : MetaData("citation_doi")
        * abstract : MetaData("citation_abstract")
        * date_publication : MetaData("citation_date")

    These *can* be overwritten if required. Also, see :class:`centaurminer.MiningEngine` to see how authors specifically are handled.

    To include an :class:`Element <centaurminer.Element>` for another piece of information, just subclass this class and add a static variable that stores an :class:`Element <centaurminer.Element>`.
    '''
    # High priority
    title = MetaData("citation_title")
    authors = MetaData("citation_author")
    doi = MetaData("citation_doi")
    abstract = MetaData("citation_abstract")
    # full_body = None
    # licensing = None

    # Medium priority
    date_publication = MetaData("citation_date")
    # citations = None
    # references = None
    extra_link = MetaData("citation_pdf_url")

    # Low priority
    # organization_affiliated = None
    # category = None
    # keywords: None
    # source_impact_factor = None

    # _defaults = ['title', 'authors', 'doi', 'abstract', 'date_publication', 'extra_link']

    @classmethod
    def _elements(cls):
        '''Returns the elements which have locations defined'''
        return [v for v in dir(cls) if not v.startswith('_') and not callable(v)]



def TagList(str_list, tag="item"):
    '''
    Tags a list and converts to a string - avoids data corruption by using a more complex tag.

    Arguments
    ---------
    str_list : list
        List of strings to be joined with tags.
    tag : str, optional
        Tag to use for the list. This tag will be wrapped in `<...>` and `</...>` to close.
    '''
    result = ""
    for string in str_list:
        result += "<item>" + string + "</item>"
    return result


def CollectURLs(start_url, link_elem, next_elem=None):
    '''
    Collects a list of URLs from a search of the site.

    Arguments
    ---------
    start_url : str
        URL for the first page of a site search.
    link_elem : :class:`centaurminer.Element`
        An Element indicating where individual URL links can be found on the search page.
    next_elem : :class:`centaruminer.Element`, optional
        Indicates where on the page the "next page" button is, to navigate through search pages.
    '''
    from .DOM_elements import PageLocations
    from .Engine import MiningEngine

    urls = []
    pageNum = 1
    # Get access to a webdriver and load the page
    miner = MiningEngine(PageLocations)
    miner.wd.get(start_url)

    # Load all the links on this page
    elems = miner.get(link_elem, several=True)
    if next_elem is None:
        return elems

    # Repeatedly gather links from several pages
    while len(elems) > 0:
        print("Appending URLs from page", pageNum)
        urls.extend(elems)
        if len(urls) > 50:
            break

        # Go to next page
        try:
            print("Going to page", pageNum + 1)
            next_button = miner.wd.find_element(next_elem.method, next_elem.selector)
            next_button.click()
        except Exception:
            break
        elems = miner.get(link_elem, several=True)
        pageNum += 1
    return urls



from .DOM_elements import Element
from .DOM_elements import MetaData
from .DOM_elements import PageLocations
from .DOM_elements import Complex

from .Engine import MiningEngine
from .utils import TagList
from .utils import CollectURLs

__all__ = [
    "Element",
    "MetaData",
    "PageLocations",
    "Complex",
    "MiningEngine",
    "TagList",
    "CollectURLs"
]



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name = "centaurminer",
    version = "0.0.3",
    author = "Simon Ewing",
    author_email = "Simonsays095@gmail.com",
    description = "A selenium wrapper to help mining data from scientific literature.",
    long_description = long_description,
    long_description_content_type = "text/x-rst",
    url = "https://github.com/Simonsays095/CENTAURminer.git",
    packages = setuptools.find_packages(),
    classifiers = [
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Science/Research",
    ],
    python_requires = '>=3.4',
    install_requires = [
        'selenium',
        'webdriver_manager'
    ],
    project_urls = {
        'Documentation': 'https://centaurminer.readthedocs.io/',
        'Source code': 'https://github.com/Simonsays095/CENTAURminer'
    }
)
